from typing import Any, Callable, List, Optional, Tuple, Union

from numpy import ndarray
from pandas import DataFrame
import numpy as np
import pandas as pd

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from shap import (
        Explainer,
        KernelExplainer,
        LinearExplainer,
        SamplingExplainer,
    )
    from shap.maskers import Masker
    from shap.models import Model
    from sklearn.base import BaseEstimator
    import shap
except ModuleNotFoundError:
    Explainer = Any
    KernelExplainer = Any
    LinearExplainer = Any
    SamplingExplainer = Any
    spmatrix = Any
    Masker = Any
    Model = Any
    BaseEstimator = Any

# ------ CONSTRUCTORS -------


class shap_explainer(ExternalOpImplementation):
    _transform_id = "shap.SHAP_EXPLAINER"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[Callable, BaseEstimator],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="masker",
            annotation=Optional[Union[Callable, ndarray, DataFrame]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="link",
            annotation=Optional[Callable],
            default=None,
        ),
        SarusParameter(
            name="algorithm",
            annotation=str,
            default="auto",
        ),
        SarusParameter(
            name="output_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="linearize_link",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="seed",
            annotation=Optional[int],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["link"] is None:
            del kwargs["link"]
        return shap.Explainer(**kwargs)


# ------ SHAP EXPLAINER METHODS ------


class shap_save(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SAVE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explainer,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="out_file",
            annotation=Any,
            condition=STATIC,
        ),
        SarusParameter(
            name="model_saver",
            annotation=str,
            default=".save",
        ),
        SarusParameter(
            name="masker_saver",
            annotation=str,
            default=".save",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.save(**kwargs)


class shap_load(ExternalOpImplementation):
    _transform_id = "shap.SHAP_LOAD"
    _signature = SarusSignature(
        SarusParameter(
            name="in_file",
            annotation=Any,
            condition=STATIC,
        ),
        SarusParameter(
            name="model_loader",
            annotation=Callable,
            default=Any,
            condition=STATIC,
        ),
        SarusParameter(
            name="masker_loader",
            annotation=Callable,
            default=Any,
            condition=STATIC,
        ),
        SarusParameter(
            name="instantiate",
            annotation=bool,
            default=True,
            condition=STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.Explainer.load(**kwargs)


class SHAP_explain_row(ExternalOpImplementation):
    _transform_id = "shap.SHAP_EXPLAIN_ROW"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explainer,
            condition=STATIC | DATASPEC,
        ),
        SarusParameter(
            name="row_args",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="max_evals",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="main_effects",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="error_bounds",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="batch_size",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="outputs",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="silent",
            annotation=bool,
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.explain_row(**kwargs)


class shap_shap_values(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SHAP_VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.shap_values(**kwargs)


class shap_call(ExternalOpImplementation):
    _transform_id = "shap.SHAP_CALL"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[ndarray, DataFrame],
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        return this(kwargs["X"])


# ------ CONSTRUCTORS TREE -------


class shap_tree(ExternalOpImplementation):
    _transform_id = "shap.SHAP_TREE"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="data",
            annotation=Optional[Union[ndarray, DataFrame]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="model_output",
            annotation=str,
            default='raw',
            condition=STATIC,
        ),
        SarusParameter(
            name="feature_perturbation",
            annotation=str,
            default='interventional',
            condition=STATIC,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
            condition=STATIC,
        ),
        SarusParameter(
            name="approximate",
            annotation=bool,
            default=False,
            condition=STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.explainers.Tree(**kwargs)


# ------ CONSTRUCTORS GPUTREE -------


class shap_gputree(ExternalOpImplementation):
    _transform_id = "shap.SHAP_GPUTREE"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="data",
            annotation=Optional[Union[np.ndarray, pd.DataFrame]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="model_output",
            annotation=str,
            default='raw',
        ),
        SarusParameter(
            name="feature_perturbation",
            annotation=str,
            default='interventional',
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="approximate",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.explainers.GPUTree(**kwargs)


# ------ SHAP TREE AND GPUTREE METHODS ------


class shap_tree_shap_values(ExternalOpImplementation):
    _transform_id = "shap.SHAP_TREE_SHAP_VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="tree_limit",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="approximate",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="check_additivity",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="from_call",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.shap_values(**kwargs)


class shap_tree_interaction_values(ExternalOpImplementation):
    _transform_id = "shap.SHAP_TREE_SHAP_INTERACTION_VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="tree_limit",
            annotation=Optional[int],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.shap_interaction_values(**kwargs)


# ------ CONSTRUCTORS KERNEL -------


class shap_kernel(ExternalOpImplementation):
    _transform_id = "shap.SHAP_KERNEL"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Callable,
            condition=DATASPEC | STATIC,
            predicate=lambda x: isinstance(x, str),
        ),
        SarusParameter(
            name="data",
            annotation=Union[np.ndarray, pd.DataFrame, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="link",
            annotation=Any,
            condition=STATIC,
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["link"] is None:
            del kwargs["link"]
        return shap.KernelExplainer(**kwargs)


# ------ SHAP KERNEL METHODS ------


class shap_run(ExternalOpImplementation):
    _transform_id = "shap.SHAP_RUN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=KernelExplainer,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.run()


class shap_allocate(ExternalOpImplementation):
    _transform_id = "shap.SHAP_ALLOCATE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=KernelExplainer,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.allocate()


class shap_solve(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SOLVE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=KernelExplainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="fraction_evaluated",
            annotation=Any,
            condition=STATIC,
        ),
        SarusParameter(
            name="dim",
            annotation=Any,
            condition=STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.solve(**kwargs)


class shap_varying_groups(ExternalOpImplementation):
    _transform_id = "shap.SHAP_VARYING_GROUPS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=KernelExplainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.varying_groups(**kwargs)


class shap_explain(ExternalOpImplementation):
    _transform_id = "shap.SHAP_EXPLAIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=KernelExplainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="incoming_instance",
            annotation=Any,
            condition=STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.explain(**kwargs)


class add_sample(ExternalOpImplementation):
    _transform_id = "shap.SHAP_ADD_SAMPLE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=KernelExplainer,
            condition=STATIC | DATASPEC,
        ),
        SarusParameter(
            name="x",
            annotation=np.array,
        ),
        SarusParameter(
            name="m",
            annotation=np.array,
        ),
        SarusParameter(
            name="w",
            annotation=float,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.addsample(**kwargs)


# ------ CONSTRUCTORS LINEAR -------


class shap_linear(ExternalOpImplementation):
    _transform_id = "shap.SHAP_LINEAR"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="masker",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="link",
            annotation=Any,
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="nsamples",
            annotation=int,
            default=1000,
        ),
        SarusParameter(
            name="feature_perturbation",
            annotation=Optional[str],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["link"] is None:
            del kwargs["link"]
        return shap.explainers.Linear(**kwargs)


# ------ CONSTRUCTORS PARTITION -------
class shap_partition(ExternalOpImplementation):
    _transform_id = "shap.SHAP_PARTITION"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="masker",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="output_names",
            annotation=Any,
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="link",
            annotation=Any,
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="nsamples",
            annotation=int,
            default=1000,
        ),
        SarusParameter(
            name="feature_perturbation",
            annotation=Optional[str],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["link"] is None:
            del kwargs["link"]
        return shap.explainers.Partition(**kwargs)


# ------ CONSTRUCTORS PERMUTATION -------


class shap_permutation(ExternalOpImplementation):
    _transform_id = "shap.SHAP_PERMUTATION"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="masker",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="output_names",
            annotation=Optional[List[str]],
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="link",
            annotation=Any,
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="linearize_link",
            annotation=bool,
            condition=STATIC,
            default=True,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="nsamples",
            annotation=int,
            default=1000,
        ),
        SarusParameter(
            name="feature_perturbation",
            annotation=Optional[str],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["link"] is None:
            del kwargs["link"]
        return shap.explainers.Permutation(**kwargs)


# ------ SHAP PERMUTATION METHODS ------


class shap_permutation_explainer_shap_values(ExternalOpImplementation):
    _transform_id = "shap.SHAP_PERMUTATION_SHAP_VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=LinearExplainer,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="npermutations",
            annotation=Optional[int],
            default=10,
        ),
        SarusParameter(
            name="main_effects",
            annotation=Optional[bool],
            default=False,
        ),
        SarusParameter(
            name="error_bounds",
            annotation=Optional[bool],
            default=False,
        ),
        SarusParameter(
            name="batch_evals",
            annotation=Optional[bool],
            default=True,
        ),
        SarusParameter(
            name="silent",
            annotation=Optional[bool],
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.shap_values(**kwargs)


# ------ CONSTRUCTORS SAMPLING -------


class shap_sampling(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SAMPLING"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="data",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.explainers.Sampling(**kwargs)


# ------ SHAP SAMPLING METHODS ------


class shap_sampling_estimate(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SAMPLING_ESTIMATE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=SamplingExplainer,
            condition=STATIC | DATASPEC,
        ),
        SarusParameter(
            name="j",
            annotation=int,
        ),
        SarusParameter(
            name="f",
            annotation=Callable,
        ),
        SarusParameter(
            name="x",
            annotation=Union[pd.Series, pd.DataFrame, np.ndarray],
        ),
        SarusParameter(
            name="X",
            annotation=Union[pd.Series, pd.DataFrame, np.ndarray],
        ),
        SarusParameter(
            name="nsamples",
            annotation=Optional[int],
            default=10,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.sampling_estimate(**kwargs)


# ------ CONSTRUCTORS EXACT -------
class shap_exact(ExternalOpImplementation):
    _transform_id = "shap.SHAP_EXACT"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="masker",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="link",
            annotation=Any,
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="linearize_link",
            annotation=bool,
            condition=STATIC,
            default=True,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            condition=STATIC,
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["link"] is None:
            del kwargs["link"]
        return shap.explainers.Exact(**kwargs)


# ------ CONSTRUCTORS ADDIDTIVE -------
class shap_additive(ExternalOpImplementation):
    _transform_id = "shap.SHAP_ADDITIVE"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="masker",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="link",
            annotation=Any,
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            condition=STATIC,
            default=None,
        ),
        SarusParameter(
            name="linearize_link",
            annotation=Optional[bool],
            condition=STATIC,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["link"] is None:
            del kwargs["link"]
        return shap.explainers.Additive(**kwargs)


# ------ CONSTRUCTORS DEEP -------
class shap_deep(ExternalOpImplementation):
    _transform_id = "shap.SHAP_DEEP"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="data",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="session",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="learning_phase_flags",
            annotation=Any,
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.explainers.Deep(**kwargs)


# ------ CONSTRUCTORS GRADIENT -------
class shap_gradient(ExternalOpImplementation):
    _transform_id = "shap.SHAP_GRADIENT"
    _signature = SarusSignature(
        SarusParameter(
            name="model",
            annotation=Union[BaseEstimator, Tuple[Any, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="data",
            annotation=Union[
                Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="session",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="batch_size",
            annotation=int,
            default=50,
        ),
        SarusParameter(
            name="local_smoothing",
            annotation=Optional[float],
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.GradientExplainer(**kwargs)



from importlib import import_module
from typing import Dict, List, Optional, Type, cast
import inspect

from .typing import NO_TRANSFORM_ID, ExternalOpImplementation

MODULES = [
    "pandas",
    "sklearn",
    "numpy",
    "imblearn",
    "pandas_profiling",
    "skopt",
    "std",
    "xgboost",
    "shap",
]


def op_name(module_name: str, transform_id: str) -> str:
    """Extract the last part of a transform ID.

    Example: sklearn.SK_FIT -> SK_FIT.
    """
    mod_name, op_name = transform_id.split(".")
    assert module_name == mod_name
    return op_name


def valid_ops(module_name: str) -> List[ExternalOpImplementation]:
    """Get all ExternalOpImplementation from a module."""

    module = import_module(
        f"sarus_data_spec.manager.ops.processor.external.{module_name}"
    )
    members = inspect.getmembers(
        module,
        lambda __obj: inspect.isclass(__obj)
        and issubclass(__obj, ExternalOpImplementation),
    )
    ops_classes = [
        cast(Type[ExternalOpImplementation], op) for _, op in members
    ]
    ops_instances = [op_class() for op_class in ops_classes]
    valid_ops = [
        op for op in ops_instances if op.transform_id() != NO_TRANSFORM_ID
    ]
    return valid_ops


def ops_mapping(module_name: str) -> Dict[str, ExternalOpImplementation]:
    """Get all ExternalOpImplementation from a module.

    Return a Mapping (op_name -> op_implementation).
    """
    ops_mapping = {
        op_name(module_name, op.transform_id()): op
        for op in valid_ops(module_name)
    }
    return ops_mapping


ROUTING = {module_name: ops_mapping(module_name) for module_name in MODULES}

# These are lists of PEP and DP transforms. They are mappings {OP_ID:
# DOCSTRING}. The docstrings are displayed in the documentation to explain
# under which condition the op is PEP or DP.


def replace_none(x: Optional[str]) -> str:
    return x if x else ""


ALL_OPS: List[ExternalOpImplementation] = sum(
    [valid_ops(module_name) for module_name in MODULES], []
)
PEP_TRANSFORMS = {
    op.transform_id(): replace_none(op.pep_kind.__doc__)
    for op in ALL_OPS
    if op.pep_kind.__doc__ != ExternalOpImplementation.pep_kind.__doc__
}

DP_TRANSFORMS = {
    op.transform_id(): replace_none(
        cast(ExternalOpImplementation, op.dp_equivalent()).is_dp.__doc__
    )
    for op in ALL_OPS
    if op.dp_equivalent_id() is not None
}



from __future__ import annotations

from typing import Collection, Optional, Protocol
import logging

from sarus_data_spec.dataspec_validator.typing import DataspecValidator
from sarus_data_spec.storage.typing import Storage
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class DataspecRewriter(Protocol):
    def storage(self) -> Storage:
        ...

    def dataspec_validator(self) -> DataspecValidator:
        ...

    def variant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        privacy_limit: Optional[st.PrivacyLimit],
        salt: Optional[int] = None,
    ) -> Optional[st.DataSpec]:
        ...

    def variants(self, dataspec: st.DataSpec) -> Collection[st.DataSpec]:
        """Return all variants attached to a Dataspec."""
        ...



from typing import Any, Callable, Dict, List, Literal, Optional, Union

import numpy.typing as npt

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from sklearn import preprocessing
except ModuleNotFoundError:
    spmatrix = Any


class sk_function_transformer(ExternalOpImplementation):
    _transform_id = "sklearn.SK_FUNCTION_TRANSFORMER"
    _signature = SarusSignature(
        SarusParameter(
            name="func",
            annotation=Optional[Callable],
            default=None,
        ),
        SarusParameter(
            name="inverse_func",
            annotation=Optional[Callable],
            default=None,
        ),
        SarusParameter(
            name="validate",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="accept_sparse",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="check_inverse",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="feature_names_out",
            annotation=Optional[Union[Callable, Literal["one-to-one"]]],
            default=None,
        ),
        SarusParameter(
            name="kw_args",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="inv_kw_args",
            annotation=Optional[Dict],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return preprocessing.FunctionTransformer(**kwargs)


class sk_onehot(ExternalOpImplementation):
    _transform_id = "sklearn.SK_ONEHOT"
    _signature = SarusSignature(
        SarusParameter(
            name="categories",
            annotation=Union[Literal["auto"], List[npt.ArrayLike]],
            default="auto",
        ),
        SarusParameter(
            name="drop",
            annotation=Optional[
                Union[Literal["first", "if_binary"], npt.ArrayLike]
            ],
            default=None,
        ),
        SarusParameter(
            name="sparse",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="sparse_output",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="dtype",
            annotation=npt.DTypeLike,
            default=float,
        ),
        SarusParameter(
            name="handle_unknown",
            annotation=Literal["error", "ignore", "infrequent_if_exist"],
            default="error",
        ),
        SarusParameter(
            name="min_frequency",
            annotation=Optional[Union[int, float]],
            default=None,
        ),
        SarusParameter(
            name="max_categories",
            annotation=Optional[int],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return preprocessing.OneHotEncoder(**kwargs)


class sk_label_encoder(ExternalOpImplementation):
    _transform_id = "sklearn.SK_LABEL_ENCODER"
    _signature = SarusSignature()

    def call(self, signature: SarusSignatureValue) -> Any:
        return preprocessing.LabelEncoder()


class sk_scale(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SCALE"
    _signature = SarusSignature(
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="axis",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="with_mean",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="with_std",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return preprocessing.scale(**kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/scalar.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%sarus_data_spec/protobuf/scalar.proto\x12\x0fsarus_data_spec\"\xd0\x0f\n\x06Scalar\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12*\n\x04spec\x18\x04 \x01(\x0b\x32\x1c.sarus_data_spec.Scalar.Spec\x12;\n\nproperties\x18\x05 \x03(\x0b\x32\'.sarus_data_spec.Scalar.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xbd\x02\n\x04Spec\x12:\n\x0btransformed\x18\x01 \x01(\x0b\x32#.sarus_data_spec.Scalar.TransformedH\x00\x12.\n\x05model\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Scalar.ModelH\x00\x12\x43\n\x0eprivacy_params\x18\x03 \x01(\x0b\x32).sarus_data_spec.Scalar.PrivacyParametersH\x00\x12\x39\n\x0brandom_seed\x18\x04 \x01(\x0b\x32\".sarus_data_spec.Scalar.RandomSeedH\x00\x12\x41\n\x0fsynthetic_model\x18\x05 \x01(\x0b\x32&.sarus_data_spec.Scalar.SyntheticModelH\x00\x42\x06\n\x04spec\x1a\xbc\x01\n\x0bTransformed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12P\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\x37.sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xd3\x08\n\x05Model\x12\x11\n\targuments\x18\x01 \x01(\x0c\x12\x17\n\x0fnamed_arguments\x18\x02 \x01(\x0c\x12=\n\x0bmodel_class\x18\x03 \x01(\x0e\x32(.sarus_data_spec.Scalar.Model.ModelClass\"\xde\x07\n\nModelClass\x12\x0c\n\x08TF_KERAS\x10\x00\x12\n\n\x06SK_SVC\x10\x01\x12\r\n\tSK_ONEHOT\x10\x02\x12\n\n\x06SK_PCA\x10\x03\x12\x1b\n\x17SK_AFFINITY_PROPAGATION\x10\x04\x12\x1f\n\x1bSK_AGGLOMERATIVE_CLUSTERING\x10\x05\x12\x0c\n\x08SK_BIRCH\x10\x06\x12\r\n\tSK_DBSCAN\x10\x07\x12\x1c\n\x18SK_FEATURE_AGGLOMERATION\x10\x08\x12\r\n\tSK_KMEANS\x10\t\x12\x17\n\x13SK_MINIBATCH_KMEANS\x10\n\x12\x11\n\rSK_MEAN_SHIFT\x10\x0b\x12\r\n\tSK_OPTICS\x10\x0c\x12\x1a\n\x16SK_SPECTRAL_CLUSTERING\x10\r\x12\x1c\n\x18SK_SPECTRAL_BICLUSTERING\x10\x0e\x12\x1c\n\x18SK_SPECTRAL_COCLUSTERING\x10\x0f\x12\x1a\n\x16SK_ADABOOST_CLASSIFIER\x10<\x12\x19\n\x15SK_ADABOOST_REGRESSOR\x10=\x12\x19\n\x15SK_BAGGING_CLASSIFIER\x10>\x12\x18\n\x14SK_BAGGING_REGRESSOR\x10?\x12\x1c\n\x18SK_EXTRA_TREES_REGRESSOR\x10@\x12\x1d\n\x19SK_EXTRA_TREES_CLASSIFIER\x10\x41\x12#\n\x1fSK_GRADIENT_BOOSTING_CLASSIFIER\x10\x42\x12\"\n\x1eSK_GRADIENT_BOOSTING_REGRESSOR\x10\x43\x12\x17\n\x13SK_ISOLATION_FOREST\x10\x44\x12\x1f\n\x1bSK_RANDOM_FOREST_CLASSIFIER\x10\x45\x12\x1e\n\x1aSK_RANDOM_FOREST_REGRESSOR\x10\x46\x12\x1d\n\x19SK_RANDOM_TREES_EMBEDDING\x10G\x12\x1a\n\x16SK_STACKING_CLASSIFIER\x10H\x12\x19\n\x15SK_STACKING_REGRESSOR\x10I\x12\x18\n\x14SK_VOTING_CLASSIFIER\x10J\x12\x17\n\x13SK_VOTING_REGRESSOR\x10K\x12\'\n#SK_HIST_GRADIENT_BOOSTING_REGRESSOR\x10L\x12(\n$SK_HIST_GRADIENT_BOOSTING_CLASSIFIER\x10M\x12 \n\x1cSK_REPEATED_STRATIFIED_KFOLD\x10P\x12\x12\n\x0eXGB_CLASSIFIER\x10\\\x12\x14\n\x10SK_LABEL_ENCODER\x10\x62\x12\x0c\n\x08SK_KFOLD\x10\x63\x1a}\n\x11PrivacyParameters\x12?\n\x06points\x18\x01 \x03(\x0b\x32/.sarus_data_spec.Scalar.PrivacyParameters.Point\x1a\'\n\x05Point\x12\x0f\n\x07\x65psilon\x18\x01 \x01(\x01\x12\r\n\x05\x64\x65lta\x18\x02 \x01(\x01\x1a\x1b\n\nRandomSeed\x12\r\n\x05value\x18\x01 \x01(\x05\x1a\x10\n\x0eSyntheticModelb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.scalar_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SCALAR_PROPERTIESENTRY._options = None
  _SCALAR_PROPERTIESENTRY._serialized_options = b'8\001'
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._options = None
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _SCALAR._serialized_start=59
  _SCALAR._serialized_end=2059
  _SCALAR_PROPERTIESENTRY._serialized_start=215
  _SCALAR_PROPERTIESENTRY._serialized_end=264
  _SCALAR_SPEC._serialized_start=267
  _SCALAR_SPEC._serialized_end=584
  _SCALAR_TRANSFORMED._serialized_start=587
  _SCALAR_TRANSFORMED._serialized_end=775
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_start=722
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_end=775
  _SCALAR_MODEL._serialized_start=778
  _SCALAR_MODEL._serialized_end=1885
  _SCALAR_MODEL_MODELCLASS._serialized_start=895
  _SCALAR_MODEL_MODELCLASS._serialized_end=1885
  _SCALAR_PRIVACYPARAMETERS._serialized_start=1887
  _SCALAR_PRIVACYPARAMETERS._serialized_end=2012
  _SCALAR_PRIVACYPARAMETERS_POINT._serialized_start=1973
  _SCALAR_PRIVACYPARAMETERS_POINT._serialized_end=2012
  _SCALAR_RANDOMSEED._serialized_start=2014
  _SCALAR_RANDOMSEED._serialized_end=2041
  _SCALAR_SYNTHETICMODEL._serialized_start=2043
  _SCALAR_SYNTHETICMODEL._serialized_end=2059
# @@protoc_insertion_point(module_scope)



from typing import cast
import json
import typing as t

import numpy as np
import pyarrow as pa

from sarus_data_spec import Transform
from sarus_data_spec.path import Path, path
from sarus_data_spec.protobuf.typing import Protobuf
from sarus_data_spec.protobuf.utilities import from_base64, to_base64
from sarus_data_spec.transform import filter, project
import sarus_data_spec.protobuf as sp
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st


def select_rows(
    _type: st.Type, array: pa.Array, is_optional: bool = False
) -> t.Tuple[pa.Array, pa.Array]:
    """Visitor selecting columns based on the type.
    The idea is that at each level,
    the filter for the array is computed, and for the union,
    we remove the fields that we want to filter among
    the columns
    TODO: implement missing types
    """

    class RowsSelector(st.TypeVisitor):
        batch_array: pa.Array = array
        filter_indices: pa.Array

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            self.filter_indices = pa.array(
                np.ones(len(self.batch_array.to_pylist())).astype('bool')
            )

        def Unit(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            self.filter_indices = pa.array(
                np.ones(len(self.batch_array.to_pylist())).astype('bool')
            )

        def Boolean(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Bytes(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            if is_optional:
                self.filter_indices = pa.array(
                    np.ones(len(self.batch_array.to_pylist())).astype('bool')
                )
            else:
                self.filter_indices = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            # no projection, we keep all existing children
            assert len(fields) == self.batch_array.type.num_fields

            arrays_indices = [
                select_rows(
                    fields[field_name],
                    self.batch_array.flatten()[
                        self.batch_array.type.get_field_index(field_name)
                    ],
                )
                for field_name in fields.keys()
            ]

            final_index = np.logical_and.reduce(
                [array_index[1] for array_index in arrays_indices]
            )
            if is_optional:
                self.filter_indices = pa.array(final_index)
            else:
                index_nulls = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )
                self.filter_indices = pa.compute.and_(
                    pa.array(final_index), index_nulls
                )
            self.batch_array = pa.StructArray.from_arrays(
                [element[0] for element in arrays_indices],
                names=list(fields.keys()),
            )

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            new_array, filter_indices = select_rows(
                type, self.batch_array, is_optional=True
            )
            self.batch_array = new_array
            self.filter_indices = filter_indices

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            # build filter array for this level
            initial_field_selected = self.batch_array.field(
                'field_selected'
            ).to_numpy(zero_copy_only=False)

            indices = [
                initial_field_selected == field_name
                for field_name in fields.keys()
            ]
            index = np.logical_or.reduce(indices)
            union_level_filter = pa.array(index)

            # now get sub-levels filter_indices
            arrays = [
                self.batch_array.flatten()[
                    self.batch_array.type.get_field_index(field_name)
                ]
                for i, (field_name, field_type) in enumerate(fields.items())
            ]
            arrays_indices = [
                select_rows(typ, arr)
                for typ, arr in zip(fields.values(), arrays)
            ]
            low_level_indices = [
                array_index[1] for array_index in arrays_indices
            ]
            low_level_filter = np.logical_or.reduce(low_level_indices)

            final_filter = pa.compute.and_(
                low_level_filter, union_level_filter
            )
            if is_optional:
                self.filter_indices = final_filter
            else:
                index_nulls = pa.compute.invert(
                    self.batch_array.is_null(nan_is_null=True)
                )
                self.filter_indices = pa.compute.and_(
                    final_filter, index_nulls
                )

            fields_name = list(fields.keys())
            # here we need to take arrays modified from low level
            selected_arrays = [
                array_index[0] for array_index in arrays_indices
            ]
            struct_fields = [
                pa.field(name=field_name, type=arr.type)
                for arr, field_name in zip(selected_arrays, fields_name)
            ]
            struct_fields.append(
                pa.field(
                    name='field_selected', type=pa.string(), nullable=False
                )
            )
            selected_arrays.append(self.batch_array.field('field_selected'))
            self.batch_array = pa.StructArray.from_arrays(
                selected_arrays,
                fields=struct_fields,
            )

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

    visitor = RowsSelector()
    _type.accept(visitor)
    return visitor.batch_array, visitor.filter_indices


def select_columns(_type: st.Type, array: pa.Array) -> pa.Array:
    """Visitor selecting columns based on the type.
    The idea is that at each level,
    the filter for the array is computed, and for the union,
    we remove the fields that we want to filter among
    the columns. Currently, it is not possible
    to filter Optional values.
    TODO: implement missing types
    """

    class ColumnsSelector(st.TypeVisitor):
        batch_array: pa.Array = array

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            arrays = [
                select_columns(
                    fields[field_name],
                    self.batch_array.flatten()[
                        self.batch_array.type.get_field_index(field_name)
                    ],
                )
                for field_name in fields.keys()
            ]
            self.batch_array = pa.StructArray.from_arrays(
                arrays,
                names=list(fields.keys()),
            )

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.batch_array = select_columns(type, self.batch_array)

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert len(fields) == self.batch_array.type.num_fields - 1
            # -1 because field_selected is not in the fields
            arrays = [
                select_columns(
                    field_type,
                    self.batch_array.flatten()[
                        self.batch_array.type.get_field_index(field_name)
                    ].filter(
                        pa.array(
                            self.batch_array.field('field_selected').to_numpy(
                                zero_copy_only=False
                            )
                            == field_name
                        )
                    ),
                )
                for field_name, field_type in fields.items()
            ]
            sizes = [len(array) for array in arrays]

            structs = []
            for i, array in enumerate(arrays):
                structs.append(
                    pa.concat_arrays(
                        [
                            pa.array([None] * sum(sizes[:i]), type=array.type),
                            array,
                            pa.array(
                                sum(sizes[i + 1 :]) * [None], type=array.type
                            ),
                        ]
                    )
                )

            structs.append(self.batch_array.field('field_selected'))
            field_names = list(fields.keys())
            field_names.append('field_selected')
            self.batch_array = pa.StructArray.from_arrays(
                structs,
                names=field_names,
            )

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Boolean(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Bytes(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Unit(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

    visitor = ColumnsSelector()
    _type.accept(visitor)
    return visitor.batch_array


def filter_project_from_types(
    initial_type: st.Type, goal_type: st.Type
) -> t.Tuple[Transform, Transform]:
    """Creates the filtering and projection to be executed
    in that order to obtain the goal type from the initial type"""

    filter_type = sdt.extract_filter_from_types(initial_type, goal_type)
    return filter(filter=filter_type), project(projection=goal_type)


def filter_primary_keys(old_pks: str, new_type: st.Type) -> str:
    """Keeps only primary keys path that appear in new type"""
    filtered_pks = []
    primary_keys = [
        Path(cast(sp.Path, from_base64(proto, cast(Protobuf, sp.Path))))
        for proto in json.loads(old_pks)
    ]
    for primary_key in primary_keys:
        try:
            new_type.get(primary_key)
        except KeyError:
            pass

        else:
            filtered_pks.append(to_base64(primary_key.protobuf()))
    return json.dumps(filtered_pks)


def update_fks(curr_type: st.Type, original_type: st.Type) -> st.Type:
    """TODO: implement missing types"""

    class Select(st.TypeVisitor):
        result = curr_type

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            new_fields = {}
            for fieldname, fieldtype in fields.items():
                new_fields[fieldname] = update_fks(
                    curr_type=fieldtype, original_type=original_type
                )

            self.result = sdt.Struct(
                fields=new_fields,
                name=name if name is not None else 'Struct',
                properties=properties,
            )
            # otherwise struct is empty and it is a terminal node

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            new_fields = {}
            for fieldname, fieldtype in fields.items():
                new_fields[fieldname] = update_fks(
                    curr_type=fieldtype, original_type=original_type
                )

            self.result = sdt.Union(
                fields=new_fields,
                name=name if name is not None else 'Union',
                properties=properties,
            )

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.result = sdt.Optional(
                type=update_fks(curr_type=type, original_type=original_type),
                name=name if name is not None else 'Optional',
                properties=properties,
            )

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.result = sdt.List(
                type=update_fks(curr_type=type, original_type=original_type),
                max_size=max_size,
                name=name if name is not None else 'Optional',
                properties=properties,
            )

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.result = sdt.Array(
                type=update_fks(curr_type=type, original_type=original_type),
                shape=shape,
                name=name if name is not None else 'Optional',
                properties=properties,
            )

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if reference is not None:
                try:
                    original_type.get(path(paths=[reference]))
                except KeyError:
                    self.result = sdt.Id(unique=unique, base=base)

        def Boolean(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Bytes(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Unit(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

    visitor = Select()
    curr_type.accept(visitor)
    return visitor.result



import pyarrow as pa

import sarus_data_spec.typing as st


def convert_record_batch(
    record_batch: pa.RecordBatch, _type: st.Type
) -> pa.Array:
    if str(_type.protobuf().WhichOneof('type')) not in ['struct', 'union']:
        return record_batch.column(0)
    record_schema = record_batch.schema
    fields = [record_schema.field(i) for i in range(len(record_schema.types))]
    return pa.StructArray.from_arrays(record_batch.columns, fields=fields)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/relation.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'sarus_data_spec/protobuf/relation.proto\x12\x0fsarus_data_spec\"\xc4\x01\n\x08Relation\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x11\n\targuments\x18\x03 \x03(\t\x12\x17\n\x0fnamed_arguments\x18\x04 \x03(\t\x12=\n\nproperties\x18\x05 \x03(\x0b\x32).sarus_data_spec.Relation.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.relation_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _RELATION_PROPERTIESENTRY._options = None
  _RELATION_PROPERTIESENTRY._serialized_options = b'8\001'
  _RELATION._serialized_start=61
  _RELATION._serialized_end=257
  _RELATION_PROPERTIESENTRY._serialized_start=208
  _RELATION_PROPERTIESENTRY._serialized_end=257
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
from subprocess import check_call
import pathlib

from setuptools import setup
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop


def generate_proto_code():
    proto_path = "sarus_data_spec/protobuf"
    proto_it = pathlib.Path().glob(proto_path + "/**/*.proto")
    protos = [str(proto) for proto in proto_it if proto.is_file()]
    check_call(
        [
            "protoc",
            "--python_out",
            ".",
            "--mypy_out",
            ".",
            "--proto_path",
            ".",
        ]
        + protos
    )


class CompileProtobufDevelop(develop):
    """Wrapper to add protobuf compilation to run before editable package installation."""

    uninstall = False

    def run(self):
        develop.run(self)

    def install_for_development(self):
        develop.install_for_development(self)
        generate_proto_code()


class CompileProtobufBuild(build_py):
    """Wrapper to add protobuf compilation to run before package installation."""

    def run(self):
        generate_proto_code()
        build_py.run(self)


if __name__ == '__main__':
    setup(version='3.5.15')



from __future__ import annotations

from os.path import basename
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Collection,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
from urllib.parse import urlparse
import json
import typing as t
import warnings

import pandas as pd
import pyarrow as pa

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # Warning is displayed by typing.py

try:
    from sqlalchemy.engine import make_url
except ModuleNotFoundError:
    warnings.warn('SqlAlchemy not found, sql operations not available')


from sarus_data_spec.base import Referring
from sarus_data_spec.constants import DATASET_SLUGNAME
from sarus_data_spec.protobuf.utilities import to_base64
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.transform import Transform
import sarus_data_spec.protobuf as sp
import sarus_data_spec.transform as sdtr
import sarus_data_spec.typing as st

if TYPE_CHECKING:
    from sarus_data_spec.bounds import Bounds
    from sarus_data_spec.links import Links
    from sarus_data_spec.marginals import Marginals
    from sarus_data_spec.multiplicity import Multiplicity
    from sarus_data_spec.size import Size


class Dataset(Referring[sp.Dataset]):
    """A python class to describe datasets"""

    def __init__(self, protobuf: sp.Dataset, store: bool = True) -> None:
        if protobuf.spec.HasField("transformed"):
            transformed = protobuf.spec.transformed
            self._referred = {
                transformed.transform,
                *transformed.arguments,
                *list(transformed.named_arguments.values()),
            }

        super().__init__(protobuf=protobuf, store=store)

    def prototype(self) -> Type[sp.Dataset]:
        """Return the type of the underlying protobuf."""
        return sp.Dataset

    def name(self) -> str:
        return self._protobuf.name

    def doc(self) -> str:
        return self._protobuf.doc

    def is_transformed(self) -> bool:
        """Is the dataset composed."""
        return self._protobuf.spec.HasField('transformed')

    def is_file(self) -> bool:
        """Is the dataset composed."""
        return self._protobuf.spec.HasField('file')

    def is_synthetic(self) -> bool:
        """Is the dataset synthetic."""
        return self.manager().dataspec_validator().is_synthetic(self)

    def has_admin_columns(self) -> bool:
        return self.schema().has_admin_columns()

    def is_protected(self) -> bool:
        return self.schema().is_protected()

    def is_pep(self) -> bool:
        """Is the dataset PEP."""
        return self.pep_token() is not None

    def pep_token(self) -> Optional[str]:
        """Returns the dataset PEP token."""
        return self.manager().dataspec_validator().pep_token(self)

    def is_dp(self) -> bool:
        """Is the dataspec the result of a DP transform"""
        return self.manager().dataspec_validator().is_dp(self)

    def is_public(self) -> bool:
        """Is the dataset public."""
        return self.manager().dataspec_validator().is_public(self)

    def is_remote(self) -> bool:
        """Is the dataspec a remotely defined dataset."""
        return self.manager().is_remote(self)

    def is_source(self) -> bool:
        """Is the dataset not composed."""
        return not self.is_transformed()

    def sql(
        self,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> Iterator[pa.RecordBatch]:
        """Executes the sql method on the dataset"""
        return self.manager().sql(self, query, dialect, batch_size)

    def sources(
        self, type_name: t.Optional[str] = sp.type_name(sp.Dataset)
    ) -> Set[st.DataSpec]:
        """Returns the set of non-transformed datasets that are parents
        of the current dataset"""
        sources = self.storage().sources(self, type_name=type_name)
        return sources

    def status(
        self, task_names: t.Optional[t.List[str]] = None
    ) -> t.Optional[st.Status]:
        """This method return a status that contains all the
        last updates for the task_names required. It returns None if
        all the tasks are missing."""

        if task_names is None:
            task_names = []
        if type(task_names) not in [list, set, tuple]:
            raise TypeError(
                f"Invalid task_names passed to dataset.status {task_names}"
            )
        last_status = self.manager().status(self)
        if last_status is None:
            return last_status
        if all([last_status.task(task) is None for task in task_names]):
            return None
        return last_status

    def schema(self) -> st.Schema:
        return self.manager().schema(self)

    async def async_schema(self) -> st.Schema:
        return await self.manager().async_schema(self)

    def size(self) -> t.Optional[Size]:
        return cast('Size', self.manager().size(self))

    def multiplicity(self) -> t.Optional[Multiplicity]:
        return cast('Multiplicity', self.manager().multiplicity(self))

    def bounds(self) -> t.Optional[Bounds]:
        return cast('Bounds', self.manager().bounds(self))

    def marginals(self) -> t.Optional[Marginals]:
        return cast('Marginals', self.manager().marginals(self))

    def links(self) -> st.Links:
        return cast('Links', self.manager().links(self))

    def transform(self) -> st.Transform:
        return cast(
            st.Transform,
            self.storage().referrable(
                self.protobuf().spec.transformed.transform
            ),
        )

    def to_arrow(self, batch_size: int = 10000) -> t.Iterator[pa.RecordBatch]:
        return self.manager().to_arrow(self, batch_size)

    async def async_to_arrow(
        self, batch_size: int = 10000
    ) -> AsyncIterator[pa.RecordBatch]:
        return await self.manager().async_to_arrow(self, batch_size)

    def to_sql(self) -> None:
        return self.manager().to_sql(self)

    def parents(
        self,
    ) -> Tuple[
        List[Union[st.DataSpec, st.Transform]],
        Dict[str, Union[st.DataSpec, st.Transform]],
    ]:
        if not self.is_transformed():
            return list(), dict()

        args_id = self._protobuf.spec.transformed.arguments
        kwargs_id = self._protobuf.spec.transformed.named_arguments

        args_parents = [
            cast(
                Union[st.DataSpec, st.Transform],
                self.storage().referrable(uuid),
            )
            for uuid in args_id
        ]
        kwargs_parents = {
            name: cast(
                Union[st.DataSpec, st.Transform],
                self.storage().referrable(uuid),
            )
            for name, uuid in kwargs_id.items()
        }

        return args_parents, kwargs_parents

    def variant(
        self,
        kind: st.ConstraintKind,
        public_context: Collection[str] = (),
        privacy_limit: Optional[st.PrivacyLimit] = None,
        salt: Optional[int] = None,
    ) -> Optional[st.DataSpec]:
        return (
            self.manager()
            .dataspec_rewriter()
            .variant(self, kind, public_context, privacy_limit, salt)
        )

    def variants(self) -> Collection[st.DataSpec]:
        return self.manager().dataspec_rewriter().variants(self)

    def private_queries(self) -> List[st.PrivateQuery]:
        """Return the list of PrivateQueries used in a Dataspec's transform.

        It represents the privacy loss associated with the current computation.

        It can be used by Sarus when a user (Access object) reads a DP dataspec
        to update its accountant. Note that Private Query objects are generated
        with a random uuid so that even if they are submitted multiple times to
        an account, they are only accounted once (ask @cgastaud for more on
        accounting).
        """
        return self.manager().dataspec_validator().private_queries(self)

    def spec(self) -> str:
        return cast(str, self._protobuf.spec.WhichOneof('spec'))

    def __iter__(self) -> Iterator[pa.RecordBatch]:
        return self.to_arrow(batch_size=1)

    def to_pandas(self) -> pd.DataFrame:
        return self.manager().to_pandas(self)

    async def async_to_pandas(self) -> pd.DataFrame:
        return await self.manager().async_to_pandas(self)

    async def async_to(
        self, kind: t.Type, drop_admin: bool = True
    ) -> st.DatasetCastable:
        """Convert a Dataset's to a Python type."""
        return await self.manager().async_to(self, kind, drop_admin)

    def to(self, kind: t.Type, drop_admin: bool = True) -> st.DatasetCastable:
        return self.manager().to(self, kind, drop_admin)

    def to_tensorflow(self) -> tf.data.Dataset:
        return self.manager().to_tensorflow(self)

    async def async_to_tensorflow(self) -> tf.data.Dataset:
        return await self.manager().async_to_tensorflow(self)

    # A Visitor acceptor
    def accept(self, visitor: st.Visitor) -> None:
        visitor.all(self)
        if self.is_transformed():
            visitor.transformed(
                self,
                cast(
                    Transform,
                    self.storage().referrable(
                        self._protobuf.spec.transformed.transform
                    ),
                ),
                *(
                    cast(Dataset, self.storage().referrable(arg))
                    for arg in self._protobuf.spec.transformed.arguments
                ),
                **{
                    name: cast(Dataset, self.storage().referrable(arg))
                    for name, arg in self._protobuf.spec.transformed.named_arguments.items()  # noqa: E501
                },
            )
        else:
            visitor.other(self)

    def foreign_keys(self) -> Dict[st.Path, st.Path]:
        """returns foreign keys of the dataset"""
        return self.manager().foreign_keys(self)

    def dot(self) -> str:
        """return a graphviz representation of the dataset"""

        class Dot(st.Visitor):
            visited: Set[st.DataSpec] = set()
            nodes: Dict[str, Tuple[str, str]] = {}
            edges: Dict[Tuple[str, str], str] = {}

            def transformed(
                self,
                visited: st.DataSpec,
                transform: st.Transform,
                *arguments: st.DataSpec,
                **named_arguments: st.DataSpec,
            ) -> None:
                if visited not in self.visited:
                    if visited.prototype() == sp.Dataset:
                        self.nodes[visited.uuid()] = (
                            visited.name(),
                            "Dataset",
                        )
                    else:
                        self.nodes[visited.uuid()] = (visited.name(), "Scalar")

                    if not visited.is_remote():
                        for argument in arguments:
                            self.edges[
                                (argument.uuid(), visited.uuid())
                            ] = transform.name()
                            argument.accept(self)
                        for _, argument in named_arguments.items():
                            self.edges[
                                (argument.uuid(), visited.uuid())
                            ] = transform.name()
                            argument.accept(self)
                    self.visited.add(visited)

            def other(self, visited: st.DataSpec) -> None:
                if visited.prototype() == sp.Dataset:
                    self.nodes[visited.uuid()] = (
                        visited.name(),
                        "Dataset",
                    )
                else:
                    self.nodes[visited.uuid()] = (visited.name(), "Scalar")

        visitor = Dot()
        self.accept(visitor)
        result = 'digraph {'
        for uuid, (label, node_type) in visitor.nodes.items():
            shape = "polygon" if node_type == "Scalar" else "ellipse"
            result += (
                f'\n"{uuid}" [label="{label} ({uuid[:2]})", shape={shape}];'
            )
        for (u1, u2), label in visitor.edges.items():
            result += f'\n"{u1}" -> "{u2}" [label="{label} ({uuid[:2]})"];'
        result += '}'
        return result

    def primary_keys(self) -> List[st.Path]:
        return self.manager().primary_keys(self)

    def attribute(self, name: str) -> t.Optional[st.Attribute]:
        return self.manager().attribute(name=name, dataspec=self)

    def attributes(self, name: str) -> t.List[st.Attribute]:
        return self.manager().attributes(name=name, dataspec=self)


# Builders
def transformed(
    transform: st.Transform,
    *arguments: t.Union[st.DataSpec, st.Transform],
    dataspec_type: Optional[str] = None,
    dataspec_name: Optional[str] = None,
    **named_arguments: t.Union[st.DataSpec, st.Transform],
) -> st.DataSpec:
    if dataspec_type is None:
        dataspec_type, attach_info_callback = transform.infer_output_type(
            *arguments, **named_arguments
        )
    else:

        def attach_info_callback(ds: st.DataSpec) -> None:
            return

    if dataspec_name is None:
        dataspec_name = "Transformed"

    if dataspec_type == sp.type_name(sp.Scalar):
        output_dataspec: st.DataSpec = Scalar(
            sp.Scalar(
                name=dataspec_name,
                spec=sp.Scalar.Spec(
                    transformed=sp.Scalar.Transformed(
                        transform=transform.uuid(),
                        arguments=(a.uuid() for a in arguments),
                        named_arguments={
                            n: a.uuid() for n, a in named_arguments.items()
                        },
                    )
                ),
            )
        )
    else:
        properties = {}
        ds_args = [
            element
            for element in arguments
            if element.type_name() == sp.type_name(sp.Dataset)
        ]
        for element in named_arguments.values():
            if element.type_name() == sp.type_name(sp.Dataset):
                ds_args.append(element)
        if len(ds_args) == 1 and DATASET_SLUGNAME in ds_args[0].properties():
            properties[DATASET_SLUGNAME] = arguments[0].properties()[
                DATASET_SLUGNAME
            ]

        output_dataspec = Dataset(
            sp.Dataset(
                name=dataspec_name,
                spec=sp.Dataset.Spec(
                    transformed=sp.Dataset.Transformed(
                        transform=transform.uuid(),
                        arguments=(a.uuid() for a in arguments),
                        named_arguments={
                            n: a.uuid() for n, a in named_arguments.items()
                        },
                    )
                ),
                properties=properties,
            )
        )

    # Add additional information to the newly created Dataspec
    # (e.g. a mock variant)
    attach_info_callback(output_dataspec)
    return output_dataspec


def file(
    format: str,
    uri: str,
    doc: str = 'A file dataset',
    properties: Optional[Mapping[str, str]] = None,
) -> Dataset:
    return Dataset(
        sp.Dataset(
            name=basename(urlparse(uri).path),
            doc=doc,
            spec=sp.Dataset.Spec(file=sp.Dataset.File(format=format, uri=uri)),
            properties=properties,
        )
    )


def csv_file(
    uri: str,
    doc: str = 'A csv file dataset',
    properties: Optional[Mapping[str, str]] = None,
) -> Dataset:
    return Dataset(
        sp.Dataset(
            name=basename(urlparse(uri).path),
            doc=doc,
            spec=sp.Dataset.Spec(file=sp.Dataset.File(format='csv', uri=uri)),
            properties=properties,
        )
    )


def files(
    name: str,
    format: str,
    uri_pattern: str,
    doc: str = 'Dataset split into files',
    properties: Optional[Mapping[str, str]] = None,
) -> Dataset:
    return Dataset(
        sp.Dataset(
            name=name,
            doc=doc,
            spec=sp.Dataset.Spec(
                files=sp.Dataset.Files(format=format, uri_pattern=uri_pattern)
            ),
            properties=properties,
        )
    )


def csv_files(
    name: str,
    uri_pattern: str,
    doc: str = 'A csv file dataset',
    properties: Optional[Mapping[str, str]] = None,
) -> Dataset:
    return Dataset(
        sp.Dataset(
            name=name,
            doc=doc,
            spec=sp.Dataset.Spec(
                files=sp.Dataset.Files(format='csv', uri_pattern=uri_pattern)
            ),
            properties=properties,
        )
    )


def sql(
    uri: str,
    tables: Optional[
        Collection[Tuple[str, str]]
    ] = None,  # pairs schema/table_name
    properties: Optional[Mapping[str, str]] = None,
) -> Dataset:
    parsed_uri = make_url(uri)
    if parsed_uri.database is None:
        name = f'{parsed_uri.drivername}_db_dataset'
    else:
        name = parsed_uri.database
    if tables is None:
        tables = []
    return Dataset(
        sp.Dataset(
            name=name,
            doc=f'Data from {uri}',
            spec=sp.Dataset.Spec(
                sql=sp.Dataset.Sql(
                    uri=uri,
                    tables=[
                        sp.Dataset.Sql.Table(
                            schema=element[0], table=element[1]
                        )
                        for element in tables
                    ],
                )
            ),
            properties=properties,
        )
    )


def mapped_sql(
    uri: str,
    mapping_sql: Mapping[st.Path, st.Path],
    schemas: Optional[Collection[str]] = None,
) -> Dataset:
    parsed_uri = make_url(uri)
    if parsed_uri.database is None:
        name = f'{parsed_uri.drivername}_db_dataset'
    else:
        name = parsed_uri.database

    serialized_mapping = json.dumps(
        {
            to_base64(original_table.protobuf()): to_base64(
                synthetic_table.protobuf()
            )
            for original_table, synthetic_table in mapping_sql.items()
        }
    )
    properties = {'sql_mapping': serialized_mapping}
    return Dataset(
        sp.Dataset(
            name=name,
            doc=f'Data from {uri}',
            spec=sp.Dataset.Spec(
                sql=sp.Dataset.Sql(
                    uri=uri,
                )
            ),
            properties=properties,
        )
    )


if t.TYPE_CHECKING:
    test_sql: st.Dataset = sql(uri='sqlite:///:memory:')
    test_file: st.Dataset = file(format='', uri='')
    test_csv_file: st.Dataset = csv_file(uri='')
    test_files: st.Dataset = files(name='', uri_pattern='', format='')
    test_csv_files: st.Dataset = csv_files(name='', uri_pattern='')
    test_transformed: st.DataSpec = transformed(
        sdtr.protect(), sql(uri='sqlite:///:memory:')
    )



import typing as t

import pandas as pd
import pyarrow as pa

from sarus_data_spec.arrow.pandas_utils import (
    convert_pandas_metadata_to_admin_columns,
    pandas_index_columns,
    remove_pandas_index_columns,
)
from sarus_data_spec.constants import DATA, PUBLIC, USER_COLUMN, WEIGHTS
import sarus_data_spec.typing as st


async def async_to_arrow_extract_admin(
    dataset: st.Dataset, batch_size: int = 10000
) -> t.Optional[t.AsyncIterator[pa.RecordBatch]]:
    """This function return an async iterator record batches of the
    admin data if there is administrative columns.
    """
    schema = await dataset.async_schema()
    if not schema.has_admin_columns():
        return None

    # Extract admin data from DATA
    batches_async_it = await dataset.async_to_arrow(batch_size)
    pe_field_names = [PUBLIC, USER_COLUMN, WEIGHTS]

    async def extract_admin_columns(
        batches_async_it: t.AsyncIterator[pa.RecordBatch],
    ) -> t.AsyncIterator[t.Tuple[pa.RecordBatch, t.Optional[pa.RecordBatch]]]:
        async for batch in batches_async_it:
            field_names = list(batch.schema.names)
            index_cols = pandas_index_columns(batch.schema)
            pe_batch = pa.RecordBatch.from_arrays(
                [
                    batch.columns[field_names.index(field_name)]
                    for field_name in pe_field_names + index_cols
                ],
                names=pe_field_names + index_cols,
            )
            pe_batch = pe_batch.replace_schema_metadata(batch.schema.metadata)
            yield pe_batch

    return extract_admin_columns(batches_async_it)


async def async_to_arrow_extract_data_only(
    dataset: st.Dataset, batch_size: int = 10000
) -> t.AsyncIterator[pa.RecordBatch]:
    """This function return an async iterator of record batches.

    The RecordBatches contain the data only, without the admin columns.
    """
    batches_async_it = await dataset.async_to_arrow(batch_size)
    schema = await dataset.async_schema()

    if not schema.has_admin_columns():
        return batches_async_it

    # Extract PE from DATA
    data_cols = list(schema.type().data_type().children().keys())

    async def extract_data(
        batches_async_it: t.AsyncIterator[pa.RecordBatch],
    ) -> t.AsyncIterator[pa.RecordBatch]:
        async for batch in batches_async_it:
            # We add the index columns to the data
            field_names = list(batch.schema.names)
            index_cols = pandas_index_columns(batch.schema)
            index_arrays = [
                batch.columns[field_names.index(col)]
                for col in pandas_index_columns(batch.schema)
            ]
            data_arrays = batch.columns[field_names.index(DATA)].flatten()
            arrays = index_arrays + data_arrays
            names = index_cols + data_cols

            if len(arrays) != len(names):
                raise ValueError(
                    f"Incompatible number of arrays {len(arrays)} and"
                    f" names {len(names)}.\n"
                    f"Names are index cols {index_cols} and data "
                    f"cols {data_cols}.\n"
                    f"There are {len(index_arrays)} index arrays "
                    f"and {len(data_arrays)} data arrays.\n"
                    f"Arrow batch schema is {batch.schema}."
                )
            new_struct_array = pa.StructArray.from_arrays(arrays, names)
            data_batch = pa.RecordBatch.from_struct_array(new_struct_array)
            data_batch = data_batch.replace_schema_metadata(
                batch.schema.metadata
            )
            yield data_batch

    return extract_data(batches_async_it)


async def async_admin_data(dataset: st.Dataset) -> t.Optional[pa.Table]:
    """Return the protected entity as a pa.Table if it exists."""
    pe_batches_async_it = await async_to_arrow_extract_admin(dataset)
    if pe_batches_async_it is None:
        return None
    pe_batches = [batch async for batch in pe_batches_async_it]
    return pa.Table.from_batches(pe_batches)


def merge_data_and_admin(
    data: pa.Table, admin_data: t.Optional[pa.Table]
) -> pa.Table:
    """Merge a protection and the data.

    If the data Table has some pandas metadata attached, we remove them before
    merging with the protected entity.
    """
    if admin_data is None:
        # TODO also wrap the data in an empty protection
        return data

    data_index_columns = pandas_index_columns(data.schema)
    if len(data_index_columns) > 0:
        # There are some pandas metadata
        assert data_index_columns == pandas_index_columns(admin_data.schema)
        data = remove_pandas_index_columns(data)

    # We merge the protected entity and data in Pyarrow
    data_arrays = [
        chunked_array.combine_chunks() for chunked_array in data.columns
    ]
    data_array = pa.StructArray.from_arrays(
        data_arrays, names=data.column_names
    )
    return admin_data.append_column(DATA, data_array)


def compute_admin_data(
    input_admin_data: pa.Table, result: st.DatasetCastable
) -> pa.Table:
    """Compute the output protected entity of an external transform."""
    # We guarantee that the data.index is a reliable way to trace how
    # the rows were rearranged
    if type(result) == pd.DataFrame:
        df = t.cast(pd.DataFrame, result)
        input_pe_df = input_admin_data.to_pandas()
        return pa.Table.from_pandas(input_pe_df.loc[df.index])
    elif type(result) == pd.Series:
        sr = t.cast(pd.Series, result)
        input_pe_df = input_admin_data.to_pandas()
        return pa.Table.from_pandas(input_pe_df.loc[sr.index])
    else:
        raise TypeError(
            f"Cannot compute the admin data for type {type(result)}"
        )


def validate_admin_data(
    admin_data: t.List[pa.Table],
) -> pa.Table:
    """Check that all admin data are equal."""
    # If we reach this part then there should be only one input admin data
    if len(admin_data) == 0:
        raise ValueError("The list of input admin data is empty.")

    pe = next(iter(admin_data), None)
    if pe is None:
        raise ValueError(
            "The dataset was infered PEP but has no input admin data"
        )

    if not all([candidate.equals(pe) for candidate in admin_data]):
        raise ValueError(
            "The dataset is PEP but has several differing input admin "
            "data values"
        )
    return pe


def create_admin_columns(table: pa.Table) -> pa.Table:
    """Isolate special columns to admin columns."""
    return convert_pandas_metadata_to_admin_columns(table)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/predicate.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/predicate.proto\x12\x0fsarus_data_spec\"\xab\x04\n\tPredicate\x12\x33\n\x06simple\x18\x01 \x01(\x0b\x32!.sarus_data_spec.Predicate.SimpleH\x00\x12\x31\n\x05inter\x18\x02 \x01(\x0b\x32 .sarus_data_spec.Predicate.InterH\x00\x12\x31\n\x05union\x18\x03 \x01(\x0b\x32 .sarus_data_spec.Predicate.UnionH\x00\x12/\n\x04\x63omp\x18\x04 \x01(\x0b\x32\x1f.sarus_data_spec.Predicate.CompH\x00\x12>\n\nproperties\x18\x05 \x03(\x0b\x32*.sarus_data_spec.Predicate.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a)\n\x06Simple\x12\x10\n\x08operator\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x1a\x37\n\x05Inter\x12.\n\npredicates\x18\x01 \x03(\x0b\x32\x1a.sarus_data_spec.Predicate\x1a\x37\n\x05Union\x12.\n\npredicates\x18\x01 \x03(\x0b\x32\x1a.sarus_data_spec.Predicate\x1a\x35\n\x04\x43omp\x12-\n\tpredicate\x18\x01 \x01(\x0b\x32\x1a.sarus_data_spec.PredicateB\x0b\n\tpredicateb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.predicate_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PREDICATE_PROPERTIESENTRY._options = None
  _PREDICATE_PROPERTIESENTRY._serialized_options = b'8\001'
  _PREDICATE._serialized_start=62
  _PREDICATE._serialized_end=617
  _PREDICATE_PROPERTIESENTRY._serialized_start=343
  _PREDICATE_PROPERTIESENTRY._serialized_end=392
  _PREDICATE_SIMPLE._serialized_start=394
  _PREDICATE_SIMPLE._serialized_end=435
  _PREDICATE_INTER._serialized_start=437
  _PREDICATE_INTER._serialized_end=492
  _PREDICATE_UNION._serialized_start=494
  _PREDICATE_UNION._serialized_end=549
  _PREDICATE_COMP._serialized_start=551
  _PREDICATE_COMP._serialized_end=604
# @@protoc_insertion_point(module_scope)



from typing import AsyncIterator
import logging
import typing as t

import pyarrow as pa

from sarus_data_spec.manager.ops.base import ScalarImplementation
import sarus_data_spec.typing as st

from .privacy_params import PrivacyParams
from .random_seed import RandomSeed

logger = logging.getLogger(__name__)
try:
    from sarus_data_spec.manager.ops.source.sql_source import SourceSQL
except ModuleNotFoundError:
    logger.info(
        "sqlalquemy not installed, source SQL operations not available."
    )

try:
    from sarus_data_spec.manager.ops.source.csv.arrow import csv_to_arrow
    from sarus_data_spec.manager.ops.source.csv.schema import csv_schema
except ModuleNotFoundError:
    logger.info("CSV package not found, source CSV operations not available.")
try:
    from sarus_data_spec.manager.ops.processor.standard.synthetic import (  # noqa: E501
        SyntheticModel,
    )
except ModuleNotFoundError:
    logger.info("Synthetic generation not available")


def get_scalar_op(scalar: st.Scalar) -> t.Type[ScalarImplementation]:
    if scalar.is_random_seed():
        return RandomSeed
    elif scalar.is_privacy_params():
        return PrivacyParams
    elif scalar.is_synthetic_model():
        return SyntheticModel
    else:
        raise NotImplementedError(f"Source scalar for {scalar}")


class SourceScalar(ScalarImplementation):
    async def value(self) -> t.Any:
        OpClass = get_scalar_op(self.scalar)
        return await OpClass(self.scalar).value()


async def source_dataset_to_arrow(
    dataset: st.Dataset, batch_size: int
) -> AsyncIterator[pa.RecordBatch]:
    if dataset.is_file():
        file_format = dataset.protobuf().spec.file.format
        if file_format == "csv":
            return csv_to_arrow(dataset, batch_size=batch_size)
        else:
            raise NotImplementedError(f"File format {file_format}")

    elif dataset.protobuf().spec.HasField('sql'):
        return await SourceSQL(dataset=dataset).to_arrow(batch_size=batch_size)
    else:
        source_type = dataset.protobuf().spec.WhichOneof('spec')
        raise NotImplementedError(f"Source {source_type}")


async def source_dataset_schema(dataset: st.Dataset) -> st.Schema:
    if dataset.protobuf().spec.HasField('sql'):
        return await SourceSQL(dataset=dataset).schema()
    elif dataset.is_file():
        file_format = dataset.protobuf().spec.file.format
        if file_format == "csv":
            return await csv_schema(dataset)
        else:
            raise NotImplementedError(f"File format {file_format}")
    else:
        raise NotImplementedError



import os

import yaml

properties_file = os.path.join(
    os.path.dirname(__file__), "privacy_properties.yaml"
)
with open(properties_file) as f:
    privacy_properties = yaml.load(f.read(), Loader=yaml.Loader)

WHITELISTED_TRANSFORMS = privacy_properties["WHITELISTED_TRANSFORMS"]

__all__ = [
    "WHITELISTED_TRANSFORMS",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/constraint.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)sarus_data_spec/protobuf/constraint.proto\x12\x0fsarus_data_spec\"\x82\x02\n\x11VariantConstraint\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61taspec\x18\x02 \x01(\t\x12\x38\n\x0f\x63onstraint_kind\x18\x03 \x01(\x0e\x32\x1f.sarus_data_spec.ConstraintKind\x12\x18\n\x10required_context\x18\x04 \x03(\t\x12\x46\n\nproperties\x18\x05 \x03(\x0b\x32\x32.sarus_data_spec.VariantConstraint.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*F\n\x0e\x43onstraintKind\x12\r\n\tSYNTHETIC\x10\x00\x12\x06\n\x02\x44P\x10\x01\x12\n\n\x06PUBLIC\x10\x02\x12\x07\n\x03PEP\x10\x03\x12\x08\n\x04MOCK\x10\x04\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.constraint_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _VARIANTCONSTRAINT_PROPERTIESENTRY._options = None
  _VARIANTCONSTRAINT_PROPERTIESENTRY._serialized_options = b'8\001'
  _CONSTRAINTKIND._serialized_start=323
  _CONSTRAINTKIND._serialized_end=393
  _VARIANTCONSTRAINT._serialized_start=63
  _VARIANTCONSTRAINT._serialized_end=321
  _VARIANTCONSTRAINT_PROPERTIESENTRY._serialized_start=272
  _VARIANTCONSTRAINT_PROPERTIESENTRY._serialized_end=321
# @@protoc_insertion_point(module_scope)



"""Utils for translating sarus schema into other sql related schemas e.g.
sqlalchemy metadata.
"""
from enum import Enum
import typing as t
import warnings

from sarus_data_spec.constants import DATA
from sarus_data_spec.manager.async_utils import sync
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa
except ModuleNotFoundError:
    warnings.warn("Sqlalchemy not installed, cannot send sql queries")

try:
    from google.cloud import bigquery

except ImportError:
    warnings.warn(
        'Google API Python client and related libraries are not installed,'
        ' cannot push dataset to bigquery'
    )
    big_query_schema_field = t.Any
else:
    big_query_schema_field = bigquery.SchemaField  # type:ignore


def to_sqlalchemy_metadata(
    _type: st.Type,
    table: sa.Table,
    col_name: t.Optional[str] = None,
    nullable: bool = False,
    primary_keys: t.Optional[t.List[str]] = None,
    foreign_keys: t.Optional[t.Dict[str, str]] = None,
    administrative_cols: t.Optional[t.Dict[str, st.Type]] = None,
) -> None:
    # TODO check that if columns have a dot are quoted by sqlalchemy
    # if not we can force them to be always quoted
    """Visitor to create sqlalchemy metadata from a sarus type"""

    if administrative_cols is None:
        administrative_cols = {}
    if primary_keys is None:
        primary_keys = []
    if foreign_keys is None:
        foreign_keys = {}

    class ToSAMetaData(st.TypeVisitor):
        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.Text, nullable=nullable, quote=True
                )
            )

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.Float, nullable=nullable, quote=True
                )
            )

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            for fieldname, fieldtype in fields.items():
                to_sqlalchemy_metadata(
                    fieldtype, table, col_name=fieldname, nullable=False
                )
            assert administrative_cols is not None
            for fieldname, fieldtype in administrative_cols.items():
                to_sqlalchemy_metadata(
                    fieldtype, table, col_name=fieldname, nullable=False
                )

            if primary_keys:
                table.append_constraint(
                    sa.PrimaryKeyConstraint(
                        *primary_keys, name=f'{table.name}_pks'
                    )
                )

            if foreign_keys:
                for key, value in foreign_keys.items():
                    table.append_constraint(
                        sa.ForeignKeyConstraint(
                            [key],
                            [value],
                        )
                    )

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise ValueError()

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            to_sqlalchemy_metadata(
                type, table, col_name=col_name, nullable=True
            )

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.DateTime, nullable=nullable, quote=True
                )
            )

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.Date, nullable=nullable, quote=True
                )
            )

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.Time, nullable=nullable, quote=True
                )
            )

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.Interval, nullable=nullable, quote=True
                )
            )

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

        def Boolean(
            self,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.Boolean, nullable=nullable, quote=True
                )
            )

        def Unit(
            self,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(col_name, sa.types.String, nullable=True, quote=True)
            )

        def Bytes(
            self,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name,
                    sa.types.LargeBinary,
                    nullable=nullable,
                    quote=True,
                )
            )

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            assert col_name
            table.append_column(
                sa.Column(
                    col_name, sa.types.NullType, nullable=nullable, quote=True
                )
            )

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            enum = Enum(name, name_values)  # type: ignore
            table.append_column(
                sa.Column(
                    col_name,
                    sa.types.Enum(enum),
                    nullable=nullable,
                    quote=True,
                )
            )

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            if base:
                if base == st.IdBase.STRING:
                    table.append_column(
                        sa.Column(
                            col_name,
                            sa.types.String(450),
                            nullable=nullable,
                            quote=True,
                        )
                    )
                elif base == st.IdBase.BYTES:
                    table.append_column(
                        sa.Column(
                            col_name,
                            sa.types.BINARY,
                            nullable=nullable,
                            quote=True,
                        )
                    )
                else:
                    table.append_column(
                        sa.Column(
                            col_name,
                            sa.types.Integer,
                            nullable=nullable,
                            quote=True,
                        )
                    )

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            if base in [st.IntegerBase.INT8, st.IntegerBase.INT16]:
                table.append_column(
                    sa.Column(
                        col_name,
                        sa.types.SmallInteger,
                        nullable=nullable,
                        quote=True,
                    )
                )
            elif base == st.IntegerBase.INT32:
                table.append_column(
                    sa.Column(
                        col_name,
                        sa.types.Integer,
                        nullable=nullable,
                        quote=True,
                    )
                )
            else:
                table.append_column(
                    sa.Column(
                        col_name,
                        sa.types.BigInteger,
                        nullable=nullable,
                        quote=True,
                    )
                )

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

    visitor = ToSAMetaData()
    _type.accept(visitor)


# https://cloud.google.com/bigquery/docs/reference/rest/v2/tables#TableFieldSchema.FIELDS.mode
class BQFieldType(Enum):
    """big query column constraint"""

    # it accepts null
    NULLABLE = 'NULLABLE'
    # id donesn't accept null values
    REQUIRED = 'REQUIRED'
    # not used for the moment
    REPEATED = 'REPEATED'


def to_bq_schema(
    _type: st.Type,
    bq_schema: t.Dict[str, big_query_schema_field],
    col_name: t.Optional[str] = None,
    typemode: t.Optional[str] = None,
) -> None:
    """Visitor that associate to each column a bigquery.SchemaField
    containing the wanted column type.
    """

    class ToBQSchema(st.TypeVisitor):
        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "STRING", mode=typemode
            )

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "FLOAT64", mode=typemode
            )

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            for fieldname, fieldtype in fields.items():
                to_bq_schema(
                    fieldtype,
                    bq_schema,
                    col_name=fieldname,
                    typemode=BQFieldType.REQUIRED.value,
                )

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise ValueError()

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            to_bq_schema(
                type,
                bq_schema,
                col_name=col_name,
                typemode=BQFieldType.NULLABLE.value,
            )

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "DATETIME", mode=typemode
            )

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "DATE", mode=typemode
            )

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "TIME", mode=typemode
            )

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

        def Boolean(
            self,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "BOOL", mode=typemode
            )

        def Unit(
            self,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            # Since big query has no type Null
            # We are considering as a nullable string
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "STRING", mode=BQFieldType.NULLABLE.value
            )

        def Bytes(
            self,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "BYTES", mode=typemode
            )

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            # treated in the same way as in Unit
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "STRING", mode=BQFieldType.NULLABLE.value
            )

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, bigquery.enums.SqlTypeNames.STRING, mode=typemode
            )

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            if base:
                if base == st.IdBase.STRING:
                    bq_schema[col_name] = bigquery.SchemaField(
                        col_name, "STRING", mode=typemode
                    )
                elif base == st.IdBase.BYTES:
                    bq_schema[col_name] = bigquery.SchemaField(
                        col_name, "BYTES", mode=typemode
                    )
                else:
                    bq_schema[col_name] = bigquery.SchemaField(
                        col_name, "INTEGER", mode=typemode
                    )

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            assert col_name
            assert typemode
            bq_schema[col_name] = bigquery.SchemaField(
                col_name, "INT64", mode=typemode
            )

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError()

    visitor = ToBQSchema()
    _type.accept(visitor)


async def async_sa_metadata_from_dataset(dataset: st.Dataset) -> sa.MetaData:
    metadata = sa.MetaData()
    manager = dataset.manager()
    schema = await manager.async_schema(dataset)
    tables = schema.tables()
    schema_type = schema.type()
    if schema_type.has_admin_columns():
        additional_cols = schema_type.children()
        additional_cols.pop(DATA)

    else:
        additional_cols = {}
    for table_path in tables:
        explicit_table_path = table_path.to_strings_list()[0]
        qualified_tablename = (
            explicit_table_path[1:]
            if explicit_table_path[0] == DATA
            else explicit_table_path
        )
        table_type = schema.data_type().sub_types(table_path)[0]
        table_name = ".".join([f'"{name}"' for name in qualified_tablename])
        # we create a fully qualified name for the table
        # it also needs to be quoted in order to deal with dots in names
        # with allows that tables in the rendered queries are fully
        # qualified to avoid amibous names.
        sa_tabmetadata = sa.Table(table_name, metadata, quote=False)
        to_sqlalchemy_metadata(
            table_type, sa_tabmetadata, administrative_cols=additional_cols
        )
    return metadata


def sa_metadata_from_dataset(dataset: st.Dataset) -> t.Any:
    return sync(async_sa_metadata_from_dataset(dataset))
